import json
import pydantic
import typing
import zipfile


class DataPointModel(pydantic.BaseModel, extra=pydantic.Extra.ignore):
    """A Pydantic validator class that ignores unknown fields,
    and fails on missing fields of a given datapoint."""

    id: int
    created_at: str
    lang: str
    text: str
    # enriched fields
    # generated at runtime by pipeline tasks
    # tranform placeholder
    text_clean: typing.Optional[str] = None
    # geocode placeholder
    place: typing.Optional[dict] = None
    # annotators placeholder
    annotation: typing.Optional[dict] = None


class ZipFileModel:
    """A class representation to handle input zip files"""

    def __init__(self, path: str):
        self.path = path

    def is_valid(self) -> bool:
        """Return False is path is not a zip file or it does not exist."""
        return zipfile.is_zipfile(self.path)

    def iter_bytes(self) -> typing.Iterable[bytes]:
        """Iterate content of extracted files from a zip file,
        one line at the time."""
        with zipfile.ZipFile(self.path, "r") as archive:
            for zip_ext_file in archive.infolist():
                with archive.open(zip_ext_file) as content:
                    yield from content

    def iter_jsonl(self) -> typing.Iterable[dict]:
        """Iterate content of extracted files from a zip file,
        one line at the time but converted to JSON."""
        for json_bytes in self.iter_bytes():
            try:
                yield json.loads(json_bytes)
            except json.decoder.JSONDecodeError:
                # empty dict at failure
                yield None

    def cache(self, output_path: str, jsonl_batch_gen: typing.Iterable[str]) -> None:
        """Cache processed NDJSON batches to zip file and collect metrics."""
        with zipfile.ZipFile(output_path, "w") as zf:
            for batch_id, batch_jsonl in enumerate(jsonl_batch_gen, start=1):
                # write ndjson batch to zip file
                zf.writestr("{}.ndjson".format(batch_id), batch_jsonl)




"""
ElasticSearch Client and Index Templates
"""

import datetime
import json
import typing
import requests


# ElasticSearch Index Templates for SMDRM Data Model
explicit_component_template = {
    "template": {
        "mappings": {
            "properties": {
                "annotations": {
                    "type": "nested",
                    "include_in_parent": True,
                    "properties": {
                        "annotation_type": {
                           "type": "keyword"
                        },
                        "annotation_prob": {
                            "type": "float"
                        },
                        "sanitized_text": {
                            "type": "text"
                        }
                    }
                },
                "created_at": {
                    "type": "date",
                    # Twitter date format
                    "format": "EEE LLL dd HH:mm:ss Z yyyy"
                },
                "id": {
                    "type": "text"
                },
                "lang": {
                    "type": "keyword"
                },
                "text": {
                    "type": "text"
                },
                "latitude": {
                    "type": "float"
                },
                "longitude": {
                    "type": "float"
                },
                "place_name": {
                    "type": "keyword"
                },
                "place_type": {
                    "type": "keyword"
                },
                "uploaded_at": {
                    "type": "date",
                    # https://www.elastic.co/guide/en/elasticsearch/reference/current/mapping-date-format.html#built-in-date-formats
                    "format": "strict_date_optional_time"
                },
            }
        }
    }
}


class ElasticSearchClient:
    """
    SMFR ElasticSearch Client to interact with the nodes.
    Usage:

    Initialize the ElasticSearch client:
      es = ElasticSearchClient("http://localhost:9200")
    The __init__() method sets the index, component and the index templates used at index creation.
      es.index
      es._component_id
      es._component_template
      es._index_template
    You can create the ElasticSearch index:
      es.create_index()
    The create_index() method will create a reusable component template and index template
    Now, any created index that matches the index patter will share the created template.
    """

    def __init__(self, url: str, index_prefix: str = "smdrm", component_id: str = None, component_template: dict = None, shards: int = 3, replicas: int = 0):
        # format: http://host:port
        self.url = url
        # index id
        self.index = "{prefix}_{date}".format(prefix=index_prefix, date=datetime.datetime.today().date().isoformat())
        # component template
        self._component_id = component_id or "explicit_component_template"
        self._component_template = component_template or explicit_component_template
        # index template
        self._index_template = {
            "index_patterns": ["{}_*".format(index_prefix), ],
            "template": {
                "settings": {
                    "number_of_shards": shards,
                    "number_of_replicas": replicas,
                },
                "mappings": {
                    "_source": {
                        "enabled": True
                    },
                },
            },
            "priority": 1,
            "composed_of": [self._component_id],
            "version": 1,
        }

    def __str__(self):
        return "ElasticSearchClient(url={url}, index={index})".format(url=self.url, index=self.index)

    def build_component_template(self, component_id, component_mapping):
        """
        There are two types of templates: index templates and component templates.
        Component templates are reusable building blocks that configure mappings,
        settings, and aliases. While you can use component templates to construct
        index templates, they are not directly applied to a set of indices. Index
        templates can contain a collection of component templates, as well as
        directly specify settings, mappings, and aliases.
        """
        return requests.put(
            "{url}/_component_template/{cid}".format(url=self.url, cid=component_id),
            json=component_mapping
        )

    def create_index_template(self, index_template) -> requests.Response:
        """
        An index template is a way to tell Elasticsearch how to configure an index when it is created.
        For data streams, the index template configures the stream’s backing indices as they are created.
        Templates are configured prior to index creation. When an index is created - either manually or through
        indexing a document - the template settings are used as a basis for creating the index.
        """
        return requests.put(
            "{url}/_index_template/{index}_template".format(url=self.url, index=self.index),
            json=index_template
        )

    def create_index(self) -> requests.Response:
        """Create the index."""
        self.build_component_template(self._component_id, self._component_template)
        self.create_index_template(self._index_template)
        return requests.put("{url}/{index}".format(url=self.url, index=self.index))

    def delete_index(self) -> requests.Response:
        """Delete the index."""
        return requests.delete("{url}/{index}".format(url=self.url, index=self.index))

    def doc_insert(self, data_point: dict) -> requests.Response:
        """Execute a doc insert operation to process an event in a single request.
        The document attributes are updated if the document ID exists.
        For more info, check https://www.elastic.co/guide/en/elasticsearch/reference/current/docs-index_.html.
        """
        return requests.post(
            "{url}/{index}/_doc/{id}".format(url=self.url, index=self.index, id=data_point["id"]),
            json=data_point
        )

    def bulk_insert(self, data_points: typing.List[dict]) -> requests.Response:
        """Execute a bulk insert operation to process multiple events in a single request.
        Meta defines the type of bulk action to execute on each event.
        For more info, check https://www.elastic.co/guide/en/elasticsearch/reference/current/docs-bulk.html."""
        ndjson_data = (
            "\n".join(
                "{meta}\n{event}".format(
                    meta=json.dumps({"index": {"_id": data_point["id"]}}),
                    event=json.dumps(data_point),
                )
                for data_point in data_points
            )
            + "\n"
        )
        return requests.post(
            "{url}/{index}/_bulk".format(url=self.url, index=self.index),
            headers={"Content-Type": "application/x-ndjson"},
            data=ndjson_data
        )



import os
import functools
import time
import typing


def get_version(path: str):
    """Get version from a given text file.
    File content should be a single line with the version number e.g. 1.0.0"""
    with open(path_arg(path)) as v:
        return v.read()


def path_arg(path: str) -> str:
    """Custom path argument parser."""
    if not os.path.exists(path):
        raise FileNotFoundError("Path not found.")
    if os.path.isdir(path):
        raise ValueError("Path is a directory, but a zip file is expected.")
    return os.path.abspath(path)


def iter_in_batches(generator: typing.Iterable, batch_size: int = 1000) -> typing.Iterable[list]:
    """Iterate a given generator in batches.
    Indicated for resources expensive tasks such as Annotations, and Geocoding."""
    batch = []
    for line in generator:
        batch.append(line)
        if len(batch) == batch_size:
            yield batch
            # flush batch
            batch.clear()
        continue
    # ensure all data points are yielded
    if batch:
        yield batch


def log_execution(logger):
    """Log elapsed time for successful execution, or Exception for failure for a decodated function."""
    def log(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                # track elapsed time
                time_start = time.time()
                result = f(*args, **kwargs)
                elapsed = time.time() - time_start
                logger.info(f'{f.__name__}() completed in {elapsed:.4f}s')
                return result
            except Exception as e:
                logger.exception(f'Exception in {f.__name__}(): {str(e)}')
        return wrapper
    return log



import typing


class Pipeline:
    """Pipeline Class represents a data pipeline made by a concatenation of steps.
    Each step is generator that applies a specific transformation to the stream of data."""

    def __init__(self, steps: typing.List[typing.Tuple[typing.Callable, typing.Optional[dict]]] = None):
        self.steps = steps or list()

    def add(self, step: typing.Iterable, kwargs: dict = None) -> None:
        self.steps.append((step, kwargs))

    def execute(self, datapoints: typing.Iterable[dict]) -> typing.Iterable[str]:
        iterator = datapoints
        for step, kwargs in self.steps:
            iterator = step(iterator, **kwargs) if kwargs else step(iterator)
        return iterator




"""SMDRM LIBDRM setup module."""

import os
from setuptools import find_packages, setup


def get_version():
    root = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(root, "VERSION.txt")) as v:
        version = v.read()
        return version


setup(
    name="libdrm",
    version=get_version(),
    description="Common helper modules shared by the pipeline services",
    long_description_content_type="text/markdown",
    url="https://github.com/panc86/smdrm",
    author="Emanuele Panizio",
    author_email="emanuele.panizio@ext.ec.europa.eu",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
    ],
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">3.7, <3.10",
    install_requires=[
        "pydantic~=1.8",
    ],
    project_urls={
        "Bug Reports": "https://github.com/panc86/smdrm/issues",
        "Kanban": "https://github.com/panc86/smdrm/projects/1",
        "Source": "https://github.com/panc86/smdrm",
    },
)
